from datetime import datetime

from pydantic import BaseModel, Field, RootModel


class LayerConfig(BaseModel):
    id: str
    satellite: str
    instrument: str
    measurement: str
    discipline: str
    title: str
    format: str
    colormap_id: str
    matrix_set: str
    source_name: str
    start_date: datetime
    description: str = Field(default="")
    tags: list[str] = Field(default=[])
    period: str | None = Field(default="daily")
    end_date: datetime | None = Field(default=None)
    temporal_start: datetime | None = Field(default=None)
    temporal_end: datetime | None = Field(default=None)
    date_interval: str | None = Field(default=None)


class LayerConfigs(RootModel):
    root: list[LayerConfig]



import logging
from pathlib import Path

import typer
from typing_extensions import Annotated

from worldview_config import models
from worldview_config.merge import merge_config
from worldview_config.render import render_templates

app = typer.Typer(add_completion=False, no_args_is_help=True)
state = {"verbose": False}


@app.callback()
def main(verbose: bool = False):
    """
    Commands for working with worldview configuration.
    """
    logging.basicConfig(
        level=logging.DEBUG if verbose else logging.INFO,
        format="%(levelname)0.1s %(asctime)s -- %(message)s",
    )
    if verbose:
        state["verbose"] = True


@app.command(no_args_is_help=True)
def merge(
    new_root: Annotated[
        Path,
        typer.Argument(
            help="""Root directory of the new configuration that
you wish to merge.""",
        ),
    ],
    target_root: Annotated[
        Path,
        typer.Argument(
            help="""Root directory of the existing configuration.""",
        ),
    ],
):
    merge_config(new_root=new_root, target_root=target_root)


@app.command(no_args_is_help=True)
def render(
    layer_config: Annotated[
        Path,
        typer.Argument(
            help="A JSON file containing layer configuration.",
        ),
    ],
    target: Annotated[
        Path,
        typer.Option(
            help="target path to render the templates into.",
        ),
    ] = Path.cwd(),
):
    with open(layer_config) as fp:
        layers = models.LayerConfigs.model_validate_json(fp.read())
    render_templates(target=target, layers=layers)



from collections import defaultdict
from pathlib import Path

import jinja2

from worldview_config import sub_paths, models

env = jinja2.Environment(
    loader=jinja2.FileSystemLoader(Path(__file__).parent / "templates")
)

# Worldview has different names specifically in
# the measurements section
sat_names = {
    "aqua": "Aqua",
    "terra": "Terra",
    "snpp": "Suomi NPP",
    "noaa20": "NOAA-20",
    "noaa21": "NOAA-21",
}
sensor_names = {"CRIS": "CrIS"}


def render_layer(target: Path, layer: models.LayerConfig, filetype: str):
    if filetype == "metadata":
        subdir = "metadata"
        suffix = "md"
    elif filetype == "config":
        subdir = "wv.json"
        suffix = "json"
    else:
        raise ValueError("Unexpected layer type %s", filetype)

    layer_dir = Path(
        target,
        "common",
        "config",
        subdir,
        "layers",
        layer.instrument,
        layer.satellite,
    )
    layer_dir.mkdir(parents=True, exist_ok=True)
    layer_path = layer_dir / f"{layer.id}.{suffix}"
    tmpl = env.get_template(f"layer.{suffix}.jinja")
    lp = str(layer_path)
    tmpl.stream(layer=layer.model_dump()).dump(lp)
    if filetype == "metadata":
        tmpl.stream(layer=layer.model_dump()).dump(lp.replace(".md", ".html"))
    return


def render_layer_order(target: Path, layer_ids: list[str]):
    layer_order_path = target / sub_paths["layer_order"]
    tmpl = env.get_template("layerOrder.json.jinja")
    tmpl.stream(layer_ids=layer_ids).dump(str(layer_order_path))


def get_measurements(layers: models.LayerConfigs):
    mgroup = defaultdict(lambda: defaultdict(list))
    for layer in layers.root:
        sat = sat_names.get(layer.satellite, layer.satellite)
        sensor = layer.instrument.upper()
        sensor = sensor_names.get(sensor, sensor)
        source = f"{sat}/{sensor}"
        mgroup[layer.measurement][source].append(layer.id)
    return mgroup.items()


def render_measurement(target: Path, measurement: str, sources: dict):
    measurement_dir = target / sub_paths["measurements"]
    measurement_dir.mkdir(parents=True, exist_ok=True)
    measurement_path = measurement_dir / f"{measurement}.json"
    tmpl = env.get_template("measurement.json.jinja")
    tmpl.stream(name=measurement, sources=sources).dump(str(measurement_path))


def get_disciplines(layers: models.LayerConfigs):
    disciplines = defaultdict(list)
    for layer in layers.root:
        disciplines[layer.discipline].append(layer.measurement)
    return disciplines.items()


def render_discipline(target: Path, discipline: str, measurements: list[str]):
    discipline_dir = target / sub_paths["disciplines"]
    discipline_dir.mkdir(parents=True, exist_ok=True)
    discipline_path = discipline_dir / f"{discipline.capitalize()}.json"
    tmpl = env.get_template("discipline.json.jinja")
    tmpl.stream(discipline=discipline, measurements=measurements).dump(
        str(discipline_path)
    )


def render_templates(target: Path, layers: models.LayerConfigs):
    for layer in layers.root:
        render_layer(target, layer, "config")
        render_layer(target, layer, "metadata")

    render_layer_order(target, [layer.id for layer in layers.root])

    for measurement, sources in get_measurements(layers):
        render_measurement(target, measurement, sources)

    for discipline, measurements in get_disciplines(layers):
        render_discipline(target, discipline, measurements)
    return



import os
import json
import shutil
from pathlib import Path

from worldview_config import sub_paths


def read_json(json_path: Path):
    with open(json_path) as f:
        out = json.load(f)
    return out


def merge_layer_order(target: dict, new: dict) -> dict:
    combined_layers = new["layerOrder"] + target["layerOrder"]
    combined_layers = list(set(combined_layers))
    return {"layerOrder": combined_layers}


def merge_measurement(target: dict, new: dict) -> dict:
    target = target["measurements"]
    new = new["measurements"]

    out = {"measurements": {}}
    ms = set(list(target) + list(new))
    for m in sorted(ms):
        if m in target:
            out["measurements"][m] = target[m]
        else:
            out["measurements"][m] = new[m]

        sources = set(list(target[m]["sources"]) + list(new[m]["sources"]))
        out["measurements"][m]["subtitle"] = ", ".join(sources)
        for source in sorted(sources):
            if m in target and m in new:
                if source in target[m]["sources"] and source in new[m]["sources"]:
                    out["measurements"][m]["sources"][source]["settings"] = list(
                        set(
                            target[m]["sources"][source]["settings"]
                            + new[m]["sources"][source]["settings"]
                        )
                    )
    return out


def merge_discipline(target: dict, new: dict) -> dict:
    out = target
    for d in sorted(out["categories"]["science disciplines"]):
        target_m = target["categories"]["science disciplines"][d]["measurements"]
        new_m = new["categories"]["science disciplines"][d]["measurements"]
        out["categories"]["science disciplines"][d]["measurements"] = sorted(
            list(set(target_m + new_m))
        )
    return out


def merge_json(target_path: Path, new_path: Path, category: str):
    target = read_json(target_path)
    new = read_json(new_path)
    match category:
        case "sources":
            target["sources"] |= new["sources"]
        case "features":
            target["features"] |= new["features"]
        case "layer_order":
            target = merge_layer_order(target, new)
        case "measurements":
            target = merge_measurement(target, new)
        case "disciplines":
            target = merge_discipline(target, new)
        case _:
            raise ValueError("Unsupported category %s", category)
    with open(target_path, "w") as f:
        json.dump(target, f)


def merge_config(target_root: Path, new_root: Path):
    for root, _, files in os.walk(new_root):
        for fname in files:
            new_path = Path(root, fname)
            target_path = Path(target_root, new_path.relative_to(new_root))
            for category, subpath in sub_paths.items():
                if subpath in str(new_path):
                    if target_path.exists():
                        merge_json(target_path, new_path, category)
                    else:
                        new_path.rename(target_path)
                    break
            else:
                if target_path.exists():
                    shutil.move(new_path, target_path)
                else:
                    new_path.rename(target_path)
